/* ========================================================================= *
 * BinarySearchTree interface
 * ========================================================================= */

#ifndef _BINARY_SEARCH_TREE_H_
#define _BINARY_SEARCH_TREE_H_

#include <stddef.h>
#include <stdbool.h>

/* Opaque Structure */
typedef struct tree_t BinarySearchTree;

/* ------------------------------------------------------------------------- *
 * Creates an empty BinarySearchTree (or BST).
 *
 * The BST must later be deleted by calling freeBinarySearchTree().
 *
 * RETURN
 * bst          A pointer to the BinarySearchTree, or NULL in case of error
 *
 * Note: the "void" here explicitely tells compiler there is no parameter.
 * A warning might occur without it.
 * ------------------------------------------------------------------------- */

BinarySearchTree* createEmptyBinarySearchTree(void);

/* ------------------------------------------------------------------------- *
 * Creates a balanced BinarySearchTree (or BST) with the given pairs of
 * key,value provided in two separate arrays. A key and a value are related
 * if they share the same index in the provided arrays.
 *
 * The BST must later be deleted by calling freeBinarySearchTree().
 *
 * PARAMETERS
 * keys         The keys to use in the BST, as double values
 * values       The values to store, as an array of arrays of char
 * length       Number of elements in the arrays (pre-condition: 0 < length)
 *
 * RETURN
 * bst          A pointer to the BinarySearchTree, or NULL in case of error
 * ------------------------------------------------------------------------- */

BinarySearchTree* createBinarySearchTree(const double* keys, const char** values, size_t length);

/* ------------------------------------------------------------------------- *
 * Frees the allocated memory of the given BinarySearchTree.
 *
 * PARAMETERS
 * bst          A valid pointer to a BinarySearchTree object
 *
 * NOTE
 * The element contained in the tree must not be freed !
 * ------------------------------------------------------------------------- */

void freeBinarySearchTree(BinarySearchTree* bst);

/* ------------------------------------------------------------------------- *
 * Counts the number of elements/nodes stored in the given BinarySearchTree.
 *
 * PARAMETERS
 * bst          A valid pointer to a BinarySearchTree object
 *
 * RETURN
 * nb           The amount of elements stored in bst
 * ------------------------------------------------------------------------- */

size_t getNumElements(const BinarySearchTree* bst);


/* ------------------------------------------------------------------------- *
 * Returns the tree height.
 *
 * PARAMETERS
 * bst          A valid pointer to a BinarySearchTree object
 *
 * RETURN
 * h            The height of the bst
 *              -1 if the tree is empty
 * ------------------------------------------------------------------------- */

int getHeight(const BinarySearchTree* bst);

/* ------------------------------------------------------------------------- *
 * Inserts a new pair key,value in the provided BinarySearchTree. If a node
 * with the same key already exists in the BST, its value is overriden by the
 * provided one.
 *
 * PARAMETERS
 * bst          A valid pointer to a BinarySearchTree object
 * key          The key of the new element or of the element to update
 * value        The value to store (array of char)
 *
 * RETURN
 * res          A boolean equal to true if the new element was successfully
 *              inserted, false otherwise
 * ------------------------------------------------------------------------- */

bool insertElement(BinarySearchTree* bst, double key, const char* value);

/* ------------------------------------------------------------------------- *
 * Finds an element in the provided BinarySearchTree on the basis of its key
 * and returns the associated value.
 *
 * PARAMETERS
 * bst          A valid pointer to a BinarySearchTree object
 * key          The key to search
 *
 * RETURN
 * value        The value associated to key, or NULL if the key was not found
 * ------------------------------------------------------------------------- */

const char* findElement(const BinarySearchTree* bst, double key);

/* ------------------------------------------------------------------------- *
 * Removes a key, value pair from the provided BinarySearchTree.
 *
 * PARAMETERS
 * bst          A valid pointer to a BinarySearchTree object
 * key          The key of the element to remove
 *
 * RETURN
 * value        The value associated to key, or NULL if the key was not found
 * ------------------------------------------------------------------------- */

const char* removeElement(BinarySearchTree* bst, double key);

/* ------------------------------------------------------------------------- *
 * Finds a set of elements in the provided BinarySearchTree whose the keys
 * are included in a range [key1, key2] and returns their values. The values
 * are sorted in the increasing order of the keys.
 *
 * PARAMETERS
 * bst          A valid pointer to a BinarySearchTree object
 * keyMin       Lower bound of the range (inclusive)
 * keyMax       Upper bound of the range (inclusive)
 * values       An array of values whose keys belong to the range [key1,key2],
 *              sorted to respect the increasing order of the keys
 *
 * RETURN
 * length       The length of the output array to the calling code
 *              0 if no element are in the given range. In that case,
 *              values is not initialized
 *
 * ALLOCATION ERROR
 * In case of allocation error, the length returned will be strictly greater
 * than zero but the adress pointed by values will be NULL.
 *
 *
 * Note: values is dynamically allocated. Consider freeing it, but ONLY the
 * array itself (not each string it features, because they point to those
 * which are stored in the tree).
 * ------------------------------------------------------------------------- */

size_t getElementsInRange(const BinarySearchTree* bst, double keyMin, double keyMax, const char*** values);

#endif // !_BINARY_SEARCH_TREE_H_
